/*******************************************************************************
* Copyright 2018-2022 Intel Corporation.
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

#ifndef _BLAS_HPP__
#define _BLAS_HPP__

#ifndef MKL_BLAS_COMPUTE_MODE
#define MKL_BLAS_COMPUTE_MODE compute_mode::unset
#endif

#include <CL/sycl.hpp>
#include <cstdint>
#include <complex>

#include "oneapi/mkl/types.hpp"
#include "oneapi/mkl/exceptions.hpp"

#include "oneapi/mkl/blas/types.hpp"
#include "oneapi/mkl/blas/buffer.hpp"
#include "oneapi/mkl/blas/usm.hpp"

#endif  /* _BLAS_HPP__ */
